package org.tinyjee.maven.dim.extensions;

import com.thoughtworks.qdox.model.JavaClass;
import org.tinyjee.maven.dim.spi.ResourceResolver;
import org.tinyjee.maven.dim.utils.AbstractAliasHandler;
import org.tinyjee.maven.dim.utils.JavaClassLoader;
import org.tinyjee.maven.dim.utils.PrintableMap;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedSet;

/**
 * DirectoryLoader is a 'source-class' compatible helper class that can scan and load files from the given source directories.
 * <p/>
 * <b>Usage:</b><div><code>
 * %{include|source=template.vm|source-class=org.tinyjee.maven.dim.extensions.DirectoryLoader}
 * </code></div>
 * or when using the alias:<div><code>
 * %{include|source=template.vm|source-directories=src/main,src/test}
 * </code></div>
 *
 * @author Juergen_Kellerer, 2011-10-09
 */
public class DirectoryLoader extends HashMap<String, Object> {

	// TODO for Version 1.1.1

	private static final long serialVersionUID = -3519896231202791446L;

	protected static final Map<String, Class<? extends Map<String, Object>>> EXTENSION_TO_LOADER_MAPPING =
			new HashMap<String, Class<? extends Map<String, Object>>>();
	protected static final Map<Class, String> FILE_INPUT_PARAM_MAP = new HashMap<Class, String>();

	static {
		EXTENSION_TO_LOADER_MAPPING.put(".xml", XmlLoader.class);
		EXTENSION_TO_LOADER_MAPPING.put(".xsl", XmlLoader.class);
		EXTENSION_TO_LOADER_MAPPING.put(".xdoc", XmlLoader.class);
		EXTENSION_TO_LOADER_MAPPING.put(".xhtml", XmlLoader.class);
		EXTENSION_TO_LOADER_MAPPING.put(".json", XmlLoader.class);
		EXTENSION_TO_LOADER_MAPPING.put(".java", JavaSourceLoader.class);
		EXTENSION_TO_LOADER_MAPPING.put(".properties", PropertiesLoader.class);

		FILE_INPUT_PARAM_MAP.put(XmlLoader.class, "xml");
		FILE_INPUT_PARAM_MAP.put(JavaSourceLoader.class, "java-source");
		FILE_INPUT_PARAM_MAP.put(PropertiesLoader.class, "properties");
	}

	/**
	 * Implements the "{@link org.tinyjee.maven.dim.extensions.DirectoryLoader#PARAM_ALIAS}" alias functionality.
	 */
	public static class AliasHandler extends AbstractAliasHandler {
		/**
		 * Constructs the handler (Note: Is called by {@link org.tinyjee.maven.dim.spi.RequestParameterTransformer#TRANSFORMERS}).
		 */
		public AliasHandler() {
			super(PARAM_ALIAS, PARAM_DIRECTORIES, DirectoryLoader.class.getName());
		}
	}

	/**
	 * Defines a shortcut for the request properties 'source-class' and 'directories'.
	 * <p/>
	 * If this parameter is present inside the request parameters list, the system will effectively behave as if
	 * 'source-class' and 'directories' where set separately.
	 */
	public static final String PARAM_ALIAS = "source-directories";

	/**
	 * Sets a comma separated list of URL or local path pointing to directories to load recursively.
	 * <p/>
	 * The specified directories are resolved in exactly the same way as when using the parameter "<code>source</code>"
	 * (see Macro reference).
	 */
	public static final String PARAM_DIRECTORIES = "directories";

	/**
	 * Is set with a selectable list of {@link FileReference} instances of all files that were found in the given directories.
	 * The returned list derives from {@link org.tinyjee.maven.dim.utils.SelectableArrayList}.
	 */
	public static final String PARAM_OUT_FILES = "files";
	/**
	 * Is set with a selectable list of {@link FileReference} instances of all <i>XML</i> files that were found in the given
	 * directories.
	 * Note: This is a shortcut to {@code $files.selectMatching("(?i).*(\.xml|\.xsl|\.xdoc|\.xhtml|\.json)$")}.
	 */
	public static final String PARAM_OUT_XML_FILES = "xmlFiles";
	/**
	 * Is set with a selectable list of {@link FileReference} instances of all <i>XML</i> files that were found in the given
	 * directories.
	 * Note: This is a shortcut to {@code $files.selectMatching("(?i).*(\.properties)$")}.
	 */
	public static final String PARAM_OUT_PROPERTY_FILES = "propertyFiles";
	/**
	 * Is set with a selectable list of all java classes found in the given directories.
	 * <p/>
	 * The returned list derives from {@link SelectableMappedJavaEntitiesList} and every list row is a map created by
	 * {@link JavaSourceLoader}.
	 */
	public static final String PARAM_OUT_JAVA_CLASSES = "javaClasses";

	/**
	 * Defines a interface to a
	 */
	interface FileReference {
		/**
		 * Returns the source file.
		 *
		 * @return the file the reference points at.
		 */
		File getFile();

		/**
		 * @return
		 */
		URI getFileUri();

		Object getContent();
	}

	public DirectoryLoader(File baseDir, Map<String, Object> requestParams) {
		putAll(requestParams);

		final JavaClassLoader javaClassLoader = JavaClassLoader.getInstance();
		final String[] directoriesList = String.valueOf(requestParams.get(PARAM_DIRECTORIES)).split("\\s*,\\s*");
		for (String directory : directoriesList) {
			final URL directoryUrl = ResourceResolver.findSource(baseDir, directory);
			if (!"file".equalsIgnoreCase(directoryUrl.getProtocol()))
				throw new IllegalArgumentException("Cannot handle remote directories " + directoryUrl);

			try {
				javaClassLoader.addSourceTree(new File(directoryUrl.toURI()));
			} catch (URISyntaxException e) {
				throw new RuntimeException(e);
			}
		}

		final SortedSet<JavaClass> allClasses = javaClassLoader.getClasses();
		final SelectableMappedJavaEntitiesList allSelectableClasses = new SelectableMappedJavaEntitiesList("javaClass");
		for (JavaClass javaClass : allClasses) {
			allSelectableClasses.add(new PrintableMap<String, Object>(
					new JavaSourceLoader(javaClass, requestParams), javaClass.getFullyQualifiedName()));
		}
		put(PARAM_OUT_JAVA_CLASSES, allSelectableClasses);
	}
}
