package org.tinyjee.maven.dim.spi;

import org.apache.maven.doxia.logging.Log;
import org.apache.maven.doxia.logging.SystemStreamLog;
import org.apache.maven.doxia.macro.MacroRequest;
import org.apache.maven.doxia.sink.Sink;
import org.codehaus.plexus.util.StringUtils;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.util.*;
import java.util.regex.Matcher;

import static java.util.Collections.synchronizedMap;
import static java.util.regex.Pattern.compile;

/**
 * Is a global, thread-local, recursive configuration singleton that may be used to exchange global settings between classes.
 *
 * @author Juergen_Kellerer, 2011-10-13
 */
public class Globals {

	static final String LB = System.getProperty("line.separator", "\n");

	private static final Map<Sink, Set<String>> attachedResources = synchronizedMap(new WeakHashMap<Sink, Set<String>>());

	private static final ThreadLocal<Stack<Globals>> threadLocal = new ThreadLocal<Stack<Globals>>() {
		@Override
		protected Stack<Globals> initialValue() {
			return new Stack<Globals>();
		}
	};

	public static Globals initializeGlobals(String charset, Sink sink, MacroRequest request, Log logger, File attachedIncludesPath) {
		Globals globals = new Globals(charset, sink, request, logger, attachedIncludesPath);
		threadLocal.get().push(globals);
		return globals;
	}

	public static Globals reInitializeGlobals(String charset, Sink sink, MacroRequest request, Log logger, File attachedIncludesPath) {
		threadLocal.get().pop();
		return initializeGlobals(charset, sink, request, logger, attachedIncludesPath);
	}

	public static void cleanupGlobals() {
		Stack<Globals> globalsStack = threadLocal.get();
		globalsStack.pop();
		if (globalsStack.empty()) threadLocal.remove();
	}

	/**
	 * Returns a thread local, recursion safe instance.
	 *
	 * @return a thread local, recursion safe instance.
	 */
	public static Globals getInstance() {
		Stack<Globals> globalsStack = threadLocal.get();
		return globalsStack.empty() ? new Globals(null, null, null, null, null) : globalsStack.peek();
	}

	/**
	 * Returns a safe logger.
	 *
	 * @return a safe logger.
	 */
	public static Log getLog() {
		return getInstance().getLogger();
	}

	/**
	 * Returns the relative path (from web root) to the directory hosting the attached includes.
	 *
	 * @return the relative path (from web root) to the directory hosting the attached includes.
	 */
	public static String getAttachedIncludesDirectory() {
		return System.getProperty("org.tinyjee.maven.dim.attachedIncludesDirectory", "attached-includes").replace('\\', '/');
	}

	private Charset charset = Charset.defaultCharset();
	private boolean charsetIsUserSupplied;
	private final MacroRequest request;
	private final File attachedIncludesPath;
	private final Sink sink;
	private Log logger;

	private Globals(String charset, Sink sink, MacroRequest request, Log logger, File attachedIncludesPath) {
		this.sink = sink;
		this.request = request;
		this.logger = logger;
		this.attachedIncludesPath = attachedIncludesPath;

		if (charsetIsUserSupplied = charset != null) {
			this.charset = Charset.forName(charset);
		} else {
			String inputEncoding = System.getProperty("org.tinyjee.maven.dim.project.inputEncoding");
			if (!StringUtils.isEmpty(inputEncoding))
				this.charset = Charset.forName(inputEncoding);
		}
	}

	/**
	 * Returns the charset to use when reading text.
	 *
	 * @return the charset to use when reading text.
	 */
	public Charset getCharset() {
		return charset;
	}

	/**
	 * Returns true if the user supplied a charset with the macro call.
	 *
	 * @return true if the user supplied a charset with the macro call.
	 */
	public boolean isCharsetIsUserSupplied() {
		return charsetIsUserSupplied;
	}

	/**
	 * Returns a macro logger to use for logging to the build log.
	 *
	 * @return a macro logger to use for logging to the build log.
	 */
	public Log getLogger() {
		if (logger == null) logger = new SystemStreamLog();
		return logger;
	}

	/**
	 * The current sink, may be 'null'.
	 *
	 * @return the current sink, may be 'null'.
	 */
	public Sink getSink() {
		return sink;
	}

	/**
	 * Creates a globals instance for the given sink.
	 * <p/>
	 * If sink is 'null' or the same as the current, the method returns 'this'.
	 *
	 * @param sink the sink to return a globals instance for.
	 * @return a globals instance that operates on the given sink.
	 */
	public Globals operateOnSink(Sink sink) {
		if (sink == null || sink == this.sink)
			return this;

		final Globals globals = new Globals(null, sink, request, logger, attachedIncludesPath);
		globals.charset = charset;
		globals.charsetIsUserSupplied = charsetIsUserSupplied;
		return globals;
	}

	/**
	 * Returns true if the sink produces some sort of HTML.
	 *
	 * @return true if the sink produces some sort of HTML.
	 */
	public boolean isHTMLCapableSink() {
		try {
			return sink instanceof org.apache.maven.doxia.markup.HtmlMarkup;
		} catch (NoClassDefFoundError ignored) {
			// Old API, use String-Search as fallback.
			String[] names = {"XhtmlSink", "XhtmlBaseSink"};
			for (Class<?> cls = sink.getClass(); cls != null; cls = cls.getSuperclass())
				for (String name : names)
					if (cls.getName().contains(name))
						return true;
			return false;
		}
	}

	/**
	 * The current macro request, may be 'null'.
	 *
	 * @return the current macro request, may be 'null'.
	 */
	public MacroRequest getRequest() {
		return request;
	}

	/**
	 * Returns the current base directory the Macro is operating against.
	 *
	 * @return the current base directory the Macro is operating against.
	 */
	public File getBasedir() {
		return request == null ? new File(".") : request.getBasedir();
	}

	/**
	 * Returns the file path to use for storing attached includes.
	 *
	 * @return the file path to use for storing attached includes, or 'null' if attaching includes as file is not supported.
	 */
	public File getAttachedIncludesPath() {
		return attachedIncludesPath;
	}

	/**
	 * Resolves the given path to an URL that can be fetched via {@link UrlFetcher}.
	 * <p/>
	 * Note: Under the hood this method delegates to {@link ResourceResolver#findSource(java.io.File, String)}.
	 *
	 * @param path the path to resolve.
	 * @return an URL that can be fetched.
	 */
	public URL resolvePath(String path) {
		return ResourceResolver.findSource(getBasedir(), path);
	}

	public Set<String> getAttachedResources() {
		Set<String> resources = attachedResources.get(sink);
		if (resources == null) attachedResources.put(sink, resources = new HashSet<String>());
		return resources;
	}

	/**
	 * Attempts to attach the given CSS (cascading style sheet).
	 *
	 * @param fileName   a unique name to reference the content.
	 * @param cssContent the content to attach, either in-lined or using the filename reference.
	 * @return a status whether attaching the cascading style sheets is supported by the underlying sink.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public boolean attachCss(String fileName, String cssContent) throws IOException {
		final Set<String> resources = getAttachedResources();
		if (resources.contains(fileName)) {
			getLog().debug("Will not attach a CSS of name '" + fileName + "' twice within the same output.");
			return true;
		}

		if (isHTMLCapableSink()) {
			StringBuilder html = new StringBuilder(64).append("<style type=\"text/css\">").append(LB);
			String link = attachContent(fileName, cssContent);
			if (link == null)
				html.append(cssContent);
			else
				html.append("   @import url('").append(link).append("');");
			html.append(LB).append("</style>").append(LB);

			sink.rawText(html.toString());
			resources.add(fileName);
			return true;
		}

		return false;
	}

	/**
	 * Attempts to attach the given client side script.
	 *
	 * @param fileName       a unique name to reference the script.
	 * @param scriptContent  the content to attach, either in-lined or using the filename reference.
	 * @param scriptLanguage the script language, defaults to "javascript".
	 * @param contentType    the content type, defaults to "text/javascript".
	 * @return a status whether attaching the script is supported by the underlying sink.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public boolean attachScript(String fileName, String scriptContent, String scriptLanguage, String contentType) throws IOException {
		final Set<String> resources = getAttachedResources();
		if (resources.contains(fileName)) {
			getLog().debug("Will not attach a Script of name '" + fileName + "' twice within the same output.");
			return true;
		}

		if (isHTMLCapableSink()) {
			if (scriptLanguage == null) contentType = "javascript";
			if (contentType == null) contentType = "text/javascript";

			StringBuilder html = new StringBuilder(64).append("<script " +
					"type=\"").append(contentType).append("\" " +
					"language=\"").append(scriptLanguage).append('"');

			String link = attachContent(fileName, scriptContent);
			if (link == null)
				html.append('>').append(LB).append(scriptContent).append(LB);
			else
				html.append(" src=\"").append(link).append("\">");
			html.append("</script>").append(LB);

			sink.rawText(html.toString());
			resources.add(fileName);
			return true;
		}

		return false;
	}


	/**
	 * Attempts to attach the given content as a file and returns the relative link to the content on success.
	 * <p/>
	 * Content attachment works only when building sites as site's resource directory must be resolvable to do so.
	 * <p/>
	 * If a file of the same name exists in the output directory, it will be overwritten with the content that was
	 * specified with the method call.
	 *
	 * @param fileName a unique name to reference the content.
	 * @param content  the content to attach.
	 * @return A relative link to the content or 'null' if attaching content is not supported by the underlying environment.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public String attachContent(String fileName, String content) throws IOException {
		if (attachedIncludesPath == null)
			return null;

		final Log log = getLog();
		Charset charset = getCharset(), utf8 = Charset.forName("UTF-8");

		// Handle charset for XML, it always defaults to UTF-8, overriding any user supplied charset.
		if (content.startsWith("<")) {
			Matcher matcher = compile("\\<\\?xml.*encoding=['\"]{1}(.+)['\"]{1}").matcher(content);
			if (matcher.find())
				charset = Charset.forName(matcher.group(1));
			else if (content.contains("/>") || !content.contains("<html"))
				charset = utf8;
		}

		final File path = attachedIncludesPath.getCanonicalFile();
		final File targetFile = new File(path, fileName).getCanonicalFile(), targetDirectory = targetFile.getParentFile();
		if (!targetFile.getAbsolutePath().startsWith(path.getAbsolutePath()))
			throw new IllegalArgumentException("Cannot store target file " + targetFile + " as it is not located below " + path);
		if (targetDirectory != null && !targetDirectory.isDirectory() && !targetDirectory.mkdirs()) {
			log.warn("Cannot create target directory '" + targetDirectory + "' to store content for file '" + fileName + '\'');
			if (log.isDebugEnabled()) log.debug("Content that will not be stored:" + LB + content);
		}

		CharsetEncoder charsetEncoder = getCharset().newEncoder();
		if (!charsetEncoder.canEncode(content)) {
			log.warn("The selected charset of '" + charset.name() + "' cannot be used to encode the specified content. " +
					"Will change the charset to '" + utf8 + '\'');
			charset = utf8;
			charsetEncoder = charset.newEncoder();
		}

		if (log.isDebugEnabled()) log.debug("About to store content to '" + targetFile + "' using charset '" + charset + '\'');
		final FileOutputStream outputStream = new FileOutputStream(targetFile);
		try {
			ByteBuffer byteBuffer = charsetEncoder.encode(CharBuffer.wrap(content));
			outputStream.write(byteBuffer.array(), byteBuffer.arrayOffset(), byteBuffer.remaining());
		} finally {
			outputStream.close();
		}

		getAttachedResources().add(fileName);

		final URI baseUri;
		final String attachedIncludesDirectory = getAttachedIncludesDirectory();
		if (!StringUtils.isEmpty(attachedIncludesDirectory)) {
			String pathString = path.getAbsolutePath().replace('\\', '/');
			if (pathString.endsWith(attachedIncludesDirectory))
				pathString = pathString.substring(0, pathString.length() - attachedIncludesDirectory.length());
			baseUri = new File(pathString).toURI();
		} else
			baseUri = path.toURI();

		return baseUri.relativize(targetFile.getCanonicalFile().toURI()).toASCIIString();
	}

	@Override
	public String toString() {
		return "Globals{" +
				"charset=" + charset +
				", charsetIsUserSupplied=" + charsetIsUserSupplied +
				", request=" + request +
				", attachedIncludesPath=" + attachedIncludesPath +
				", sink=" + sink +
				", logger=" + logger +
				'}';
	}
}
