package org.tinyjee.maven.dim.utils;

import com.thoughtworks.qdox.model.AbstractJavaEntity;
import com.thoughtworks.qdox.model.Annotation;
import com.thoughtworks.qdox.model.JavaClass;
import com.thoughtworks.qdox.model.Type;
import org.apache.maven.doxia.logging.Log;
import org.tinyjee.maven.dim.spi.Globals;

import java.lang.reflect.Field;
import java.util.*;
import java.util.regex.Pattern;

/**
 * Implements a base class for selectable lists of qdox based java entities (@{@link com.thoughtworks.qdox.model.AbstractJavaEntity})
 * using an AOP-like expressions syntax to select classes, methods and fields.
 * <p/>
 * Lists of this type may contain, classes, methods, fields or any other element that derives from the AbstractJavaEntity.
 * <p/>
 * Expression samples:<ul>
 * <li><code>MyClass</code></li>
 * <li><code>My*</code></li>
 * <li><code>my.package.MyClass</code></li>
 * <li><code>..MyClass</code></li>
 * <li><code>my.*.MyClass</code></li>
 * <li><code>my..MyClass</code></li>
 * <li><code>public void my..MyClass.myMethod(..)</code></li>
 * <li><code>public void my..MyClass.myMethod(java.lang.String, ..)</code></li>
 * <li><code>public void my..MyClass.myMethod(..String, ..)</code></li>
 * <li><code>..MyClass.myOtherMethod()</code></li>
 * <li><code>.myOtherMethod</code></li>
 * <li><code>@my..Annotation</code></li>
 * <li><code>@my..Annotation(..value="something"..)</code></li>
 * </ul>
 * <p/>
 * Note: The matching is implemented using regular expressions with the logic: ".." translates to ".*($|\.|\$)" and
 * "*" translates to "[^\w\.]*" while all other characters are escaped and matched as specified with one exception that
 * modifiers (e.g. public) and argument lists get special treatment to get rid of ordering problems. The following example
 * illustrates how java entities are converted to strings that are then matched with the regular expressions:
 *
 * @author Juergen_Kellerer, 2011-10-10
 */
public abstract class AbstractSelectableJavaEntitiesList<E> extends SelectableArrayList<E> {

	private static final long serialVersionUID = -1293142697836984789L;

	protected AbstractSelectableJavaEntitiesList() {
	}

	protected AbstractSelectableJavaEntitiesList(Collection<? extends E> collection) {
		super(collection);
	}

	protected abstract AbstractJavaEntity unwrap(E element);

	@Override
	@SuppressWarnings("unchecked")
	public AbstractSelectableJavaEntitiesList<E> clone() {
		return (AbstractSelectableJavaEntitiesList<E>) super.clone();
	}

	/**
	 * Selects all java entities that match the given AOP-like expression.
	 * <p/>
	 * Note: Prefixing the expression with "!" inverts the meaning.
	 *
	 * @param entityExpression an expression that matches entities contained in this list.
	 * @return all java entities that match the given AOP-like expression.
	 */
	@Override
	public AbstractSelectableJavaEntitiesList<E> selectMatching(final String entityExpression) {
		if (entityExpression.startsWith("!")) return selectNonMatching(entityExpression.substring(1));
		return (AbstractSelectableJavaEntitiesList<E>) select(new SimpleEntitySelector(entityExpression));
	}

	/**
	 * Selects all java entities that do not match the given AOP-like expression.
	 * <p/>
	 * Note: Prefixing the expression with "!" inverts the meaning.
	 *
	 * @param entityExpression an expression that matches entities contained in this list.
	 * @return all java entities that do not match the given AOP-like expression.
	 */
	@Override
	public AbstractSelectableJavaEntitiesList<E> selectNonMatching(String entityExpression) {
		if (entityExpression.startsWith("!")) return selectMatching(entityExpression.substring(1));
		return (AbstractSelectableJavaEntitiesList<E>) select(createInvertSelector(new SimpleEntitySelector(entityExpression)));
	}

	/**
	 * Selects all java entities that derive from a class that is matching the given AOP-like expression.
	 *
	 * @param classNameExpression an expression matching a super class or interface.
	 * @return all java entities that derive from a class that is matching the given AOP-like expression.
	 */
	public AbstractSelectableJavaEntitiesList<E> selectDerived(String classNameExpression) {
		return (AbstractSelectableJavaEntitiesList<E>) select(new DerivedEntitySelector(classNameExpression));
	}

	/**
	 * Selects all java entities that do not derive from a class that is matching the given AOP-like expression.
	 *
	 * @param classNameExpression an expression matching a super class or interface that must not be present in selected classes.
	 * @return all java entities that do not derive from a class that is matching the given AOP-like expression.
	 */
	public AbstractSelectableJavaEntitiesList<E> selectNonDerived(String classNameExpression) {
		return (AbstractSelectableJavaEntitiesList<E>) select(createInvertSelector(new DerivedEntitySelector(classNameExpression)));
	}

	/**
	 * Selects all java entities that are annotated with an annotation that is matching the given AOP-like expression.
	 *
	 * @param annotationNameExpression an expression matching an annotation (optionally including key & value pairs in method braces).
	 * @return all java entities that are annotated with an annotation that is matching the given AOP-like expression.
	 */
	public AbstractSelectableJavaEntitiesList<E> selectAnnotated(String annotationNameExpression) {
		return (AbstractSelectableJavaEntitiesList<E>) select(new AnnotatedEntitySelector(annotationNameExpression));
	}

	/**
	 * Selects all java entities that are not annotated with an annotation that is matching the given AOP-like expression.
	 *
	 * @param annotationNameExpression an expression matching an annotation (optionally including key & value pairs in method braces).
	 * @return all java entities that are not annotated with an annotation that is matching the given AOP-like expression.
	 */
	public AbstractSelectableJavaEntitiesList<E> selectNonAnnotated(String annotationNameExpression) {
		return (AbstractSelectableJavaEntitiesList<E>) select(createInvertSelector(new AnnotatedEntitySelector(annotationNameExpression)));
	}

	static final Field javaClassIsAnnotationField;

	static {
		try {
			javaClassIsAnnotationField = JavaClass.class.getDeclaredField("isAnnotation");
			javaClassIsAnnotationField.setAccessible(true);
		} catch (NoSuchFieldException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Returns true if the given java class is an annotation.
	 *
	 * @param javaClass the class to test.
	 * @return true if the given java class is an annotation.
	 */
	public static boolean isAnnotation(JavaClass javaClass) {
		try {
			return (Boolean) javaClassIsAnnotationField.get(javaClass);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private abstract class AbstractEntitySelector implements Selector<E> {

		final Log logger = Globals.getLog();
		final Pattern pattern;

		AbstractEntitySelector(String entityExpression) {
			pattern = entityExpression == null ? null : compileAOPLikeExpression(entityExpression);
		}

		public boolean accept(E element) {
			AbstractJavaEntity javaEntity = unwrap(element);
			return javaEntity != null && accept(javaEntity);
		}

		abstract boolean accept(AbstractJavaEntity entity);

		boolean matches(Object entity) {
			String entityString = entityToString(entity);
			boolean matches = pattern.matcher(entityString).matches();
			if (logger.isDebugEnabled()) logger.debug(pattern.pattern() + (matches ? " matches " : " does NOT match ") + entityString);
			return matches;
		}

		String entityToString(Object entity) {
			if (entity instanceof JavaClass) {
				return toString((JavaClass) entity);
			} else
				return entity.toString();
		}

		String toString(JavaClass entity) {
			StringBuilder result = new StringBuilder();
			appendModifiers(entity, result);
			result.append(entity.isEnum() ? "enum " : entity.isInterface() ? "interface " :
					isAnnotation(entity) ? "@interface " : "class ");
			result.append(entity.getFullyQualifiedName());
			return result.toString();
		}

		void appendModifiers(AbstractJavaEntity entity, StringBuilder result) {
			for (String modifier : entity.getModifiers()) result.append(modifier).append(' ');
		}

		@Override
		public String toString() {
			return getClass().getSimpleName() + '{' +
					"pattern=" + pattern +
					'}';
		}
	}

	private class SimpleEntitySelector extends AbstractEntitySelector {
		private SimpleEntitySelector(String entityExpression) {
			super(entityExpression);
		}

		@Override
		boolean accept(AbstractJavaEntity entity) {
			return matches(entity);
		}
	}


	private class DerivedEntitySelector extends AbstractEntitySelector {
		DerivedEntitySelector(String classNameExpression) {
			super(classNameExpression);
		}

		@Override
		boolean accept(AbstractJavaEntity entity) {
			if (entity instanceof JavaClass) {
				List<Type> typesToCheck = new ArrayList<Type>(Arrays.asList(((JavaClass) entity).getImplements()));
				typesToCheck.add(((JavaClass) entity).getSuperClass());

				for (Type type : typesToCheck) {
					if (type != null && (matches(type) || accept(type.getJavaClass())))
						return true;
				}
			}
			return false;
		}
	}

	private class AnnotatedEntitySelector extends AbstractEntitySelector {
		AnnotatedEntitySelector(String annotationNameExpression) {
			super(annotationNameExpression.endsWith(")") ? annotationNameExpression : annotationNameExpression + "(..)");
		}

		@Override
		boolean accept(AbstractJavaEntity entity) {
			for (Annotation annotation : entity.getAnnotations()) {
				if (matches(annotation))
					return true;
			}
			return false;
		}
	}

	static Pattern compileAOPLikeExpression(String expression) {
		final StringBuilder pattern = new StringBuilder(expression.length() + 16).append("^.*");

		// Adjust expression for annotations without package name definition.
		if (expression.startsWith("@")) {
			int packageIndex = expression.indexOf('.'), braceIndex = expression.indexOf('(');
			if (packageIndex == -1 || (braceIndex != -1 && packageIndex > braceIndex))
				expression = "@.." + expression.substring(1);
		}

		String[] modifiers = expression.split("\\s+");
		if (modifiers.length > 1) {
			// We must match all modifiers without considering the order.
			for (int j = modifiers.length - 1; j > 0; j--) {
				pattern.append('(');
				for (int i = 0, len = modifiers.length - 1; i < len; i++) {
					if (i > 0) pattern.append('|');
					pattern.append(Pattern.quote(modifiers[i]));
				}
				pattern.append(')');
			}
			pattern.append(".*");
			expression = modifiers[modifiers.length - 1];
		}

		final StringTokenizer tokenizer = new StringTokenizer(expression.replace("..", "~"), "~*()", true);
		int openBraceCount = 0;
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			switch (token.charAt(0)) {
				case '*':
					pattern.append("[^\\s\\.\\$]*");
					break;
				case '~':
					if (openBraceCount > 0)
						pattern.append(".*");
					else
						pattern.append(".*($|\\.|\\$)");
					break;
				case '(':
					openBraceCount++;
					pattern.append("\\(");
					break;
				case ')':
					openBraceCount--;
					pattern.append("\\)");
					break;
				default:
					pattern.append(Pattern.quote(token));
			}
		}
		Pattern compiledPattern = Pattern.compile(pattern.append("$").toString());

		Globals.getLog().debug("Compiled AOP like expression '" + expression +
				"' to the following regular expression: " + compiledPattern.pattern());

		return compiledPattern;
	}
}
