package org.tinyjee.maven.dim.utils;

import com.thoughtworks.qdox.JavaDocBuilder;
import com.thoughtworks.qdox.model.JavaClass;
import com.thoughtworks.qdox.model.JavaPackage;
import com.thoughtworks.qdox.model.JavaSource;
import org.tinyjee.maven.dim.spi.ResourceResolver;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.lang.ref.SoftReference;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.*;

import static java.util.Arrays.asList;

/**
 * Is a singleton that loads java classes at source level using qdox.
 *
 * @author Juergen_Kellerer, 2011-10-11
 */
public class JavaClassLoader {

	private static final JavaClassLoader instance = new JavaClassLoader();

	public static JavaClassLoader getInstance() {
		return instance;
	}

	private SoftReference<JavaDocBuilder> sharedBuilderReference = new SoftReference<JavaDocBuilder>(null);
	private final Set<String> sourceUrls = new LinkedHashSet<String>();
	private final Set<String> sourceTrees = new LinkedHashSet<String>();

	private JavaClassLoader() {
	}

	public synchronized JavaSource addSource(File basePath, String pathOrUrl) {
		return addSource(basePath, pathOrUrl, false);
	}

	public synchronized JavaSource addSource(File basePath, String pathOrUrl, boolean doNotLoadSourceTree) {
		return addSource(ResourceResolver.findSource(basePath, pathOrUrl), doNotLoadSourceTree);
	}

	public synchronized JavaSource addSource(URL sourceUrl, Reader sourceContent) {
		String sourceInfo = sourceUrl.toExternalForm();
		sourceUrls.add(sourceInfo);
		return getSharedBuilder().addSource(sourceContent, sourceInfo);
	}

	public synchronized JavaSource addSource(URL sourceUrl, boolean doNotLoadSourceTree) {
		JavaSource javaSource = getSource(sourceUrl);
		if (javaSource != null)
			return javaSource;

		sourceUrls.add(sourceUrl.toString());

		try {
			javaSource = getSharedBuilder().addSource(sourceUrl);
			if (!doNotLoadSourceTree && "file".equals(sourceUrl.getProtocol())) {
				String packagePath = javaSource.getPackageName().replace('.', '/');
				String url = sourceUrl.toString(), rootUrl = url.substring(0, url.indexOf(packagePath));
				addSourceTree(new File(new URL(rootUrl).toURI()));
			}

			return javaSource;
		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	public synchronized void addSourceTree(File sourceFolder) {
		if (sourceTrees.add(sourceFolder.toString()))
			getSharedBuilder().addSourceTree(sourceFolder);
	}

	public JavaSource getSource(URL sourceUrl) {
		for (JavaSource javaSource : getSources())
			if (sourceUrl.toString().equals(String.valueOf(javaSource.getURL())))
				return javaSource;
		return null;
	}

	public synchronized JavaSource[] getSources() {
		return getSharedBuilder().getSources();
	}

	public SortedSet<JavaClass> getClasses() {
		final JavaClass[] classes;
		synchronized (this) {
			classes = getSharedBuilder().getClasses();
		}

		return new TreeSet<JavaClass>(asList(classes));
	}

	public SortedSet<JavaPackage> getPackages() {
		final JavaPackage[] packages;
		synchronized (this) {
			packages = getSharedBuilder().getPackages();
		}

		final TreeSet<JavaPackage> javaPackages = new TreeSet<JavaPackage>(new Comparator<JavaPackage>() {
			public int compare(JavaPackage o1, JavaPackage o2) {
				return o1.getName().compareTo(o2.getName());
			}
		});
		javaPackages.addAll(asList(packages));
		return javaPackages;
	}

	public synchronized void clear() {
		sourceTrees.clear();
		sourceUrls.clear();
		sharedBuilderReference = new SoftReference<JavaDocBuilder>(null);
	}

	JavaDocBuilder getSharedBuilder() {
		JavaDocBuilder builder = sharedBuilderReference.get();
		if (builder == null)
			sharedBuilderReference = new SoftReference<JavaDocBuilder>(builder = createNewBuilder());
		return builder;
	}

	JavaDocBuilder createNewBuilder() {
		JavaDocBuilder javaDocBuilder = new JavaDocBuilder();

		// TODO: javaDocBuilder.setEncoding("utf-8");

		try {
			for (String sourceUrl : sourceUrls)
				javaDocBuilder.addSource(new URL(sourceUrl));
			for (String sourceTree : sourceTrees)
				javaDocBuilder.addSourceTree(new File(sourceTree));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

		return javaDocBuilder;
	}
}
